/*
 * Copyright (c) 2002 University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of SPAK.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation is hereby granted without fee, provided that the
 * above copyright notice and this permission/disclaimer notice is
 * retained in all copies or modified versions, and that both notices
 * appear in supporting documentation.  THE COPYRIGHT HOLDERS PROVIDE
 * THIS SOFTWARE "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE COPYRIGHT
 * HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 * RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu
 * any improvements that they make and grant redistribution rights to the
 * University of Utah.
 *
 * Author: John Regehr (regehr@cs.utah.edu)
 */

#include "spak_public.h"

#define DBG_LEVEL 1

/*
 * configuration for this module
 */

static int test_target_no_pt;
static int test_target_yes_pt;
static int minimize;
static int optimize;
static int do_sim;
int test_overrun;

/*
 * code and data for this module
 */

double scale_p, scale_np, scale_pt_4pt, scale_pt_4npt;
double scale_o_p, scale_o_np, scale_o_pt_4pt, scale_o_pt_4npt;

static void look_for_np_schedule (struct task_set *ts,
				  struct task_set **new_ts)
{
  struct task_set *ts3, *ts4;
  int feas, res3, res4;

  set_analysis (ts, "George96");  
  set_priorities (ts, INORDER);
  make_all_nonpreemptible (ts);

  ts3 = copy_task_set (ts);
  ts4 = copy_task_set (ts);
  
  res3 = assign_optimal_pri (ts3);
  
  feas = feasible (ts3, TRUE);
  assert ((res3 && (feas == num_tasks (ts3))) ||
	  (!res3 && (feas != num_tasks (ts3))));

  // print_task_set (ts3);

  set_analysis (ts4, "Wang00_fixed");
  res4 = assign_optimal_pri (ts4);
  
  feas = feasible (ts4, TRUE);
  assert ((res4 && (feas == num_tasks (ts4))) ||
	  (!res4 && (feas != num_tasks (ts4))));

  // print_task_set (ts4);
  
  assert (ensure_same_response_times (ts3, ts4));

  DBGPrint (0, ("George96 analysis: %d  / %d tasks are feasible\n", 
		feas, num_tasks (ts4)));
  
  if (feas == num_tasks (ts4)) {

    if (optimize) {
      scale_np = find_critical_scale (ts4, NULL);
#if 1
      ts4 = maximize_insensitivity_optimal (ts4);
#else
      ts4 = maximize_insensitivity_by_annealing (ts4, FALSE, FALSE);
#endif
      scale_o_np = find_critical_scale (ts4, NULL);
      DBGPrint (0, ("  non-preemptive CS: unoptimized %f,  optimized %f\n",
		    scale_np, scale_o_np));
    } else {
      scale_o_np = -1;
    }

    *new_ts = ts4;

  } else {
    *new_ts = NULL;
    scale_np = scale_o_np = -1;
    free_task_set (ts4);
  }
  DBGPrint (0, ("Wang00_fixed analysis (NP): %d  /  %d tasks are feasible\n", 
		feas, num_tasks (ts3)));

  free_task_set (ts3);
}

void look_for_p_schedule (struct task_set *ts,
			  struct task_set **new_ts)
{
  struct task_set *ts1, *ts2;
  int res1, res2;
  int feas;

  set_priorities (ts, INORDER);
  make_all_preemptible (ts);

  ts1 = copy_task_set (ts);
  ts2 = copy_task_set (ts);

  set_analysis (ts1, "Tindell92_restricted");

  res1 = assign_optimal_pri (ts1);
  
  // set_priorities (ts1, DM); // FIXME -- remove

  feas = feasible (ts1, TRUE);

  DBGPrint (0, ("Tindell92 analysis: %d / %d tasks are feasible\n", 
		feas, num_tasks (ts1)));  
  // if (res1) print_task_set (ts1);

  assert ((res1 && (feas == num_tasks (ts1))) ||
	  (!res1 && (feas != num_tasks (ts1))));

  set_analysis (ts2, "Wang00_fixed");

  res2 = assign_optimal_pri (ts2);
  
  // set_priorities (ts2, DM); // FIXME -- remove

  feas = feasible (ts2, TRUE);

  DBGPrint (0, ("Wang00 analysis (P): %d / %d tasks are feasible\n", 
		feas, num_tasks (ts2)));

  // if (res2) print_task_set (ts2);
  
  assert ((res2 && (feas == num_tasks (ts2))) ||
	  (!res2 && (feas != num_tasks (ts2))));

  assert (ensure_same_response_times (ts1, ts2));
  
  if (feas == num_tasks (ts2)) {

    if (optimize) {
      scale_p = find_critical_scale (ts2, NULL);
#if 1
      ts2 = maximize_insensitivity_optimal (ts2);
#else
      ts2 = maximize_insensitivity_by_annealing (ts2, FALSE, FALSE);
#endif
      scale_o_p = find_critical_scale (ts2, NULL);

      DBGPrint (0, ("  preemptive CS: unoptimized %f,  optimized %f\n",
		    scale_p, scale_o_p));
    } else {
      scale_o_p = -1;
    }

    *new_ts = ts2;

  } else {
    *new_ts = NULL;
    free_task_set (ts2);
    scale_p = scale_o_p = -1;
  }
  DBGPrint (0, ("Wang00 analysis (P): %d  /  %d tasks are feasible\n", 
		feas, num_tasks (ts1)));

  free_task_set (ts1);
}

static void look_for_pt_schedule (struct task_set *input_ts,
				  struct task_set **new_ts,
				  int target_has_pt)
{
  struct task_set *anneal_ts, *greedy_ts;
  int greedy_res, anneal_res;
  int feas;
  double greedy_scale_pt, greedy_scale_o_pt;
  double anneal_scale_pt, anneal_scale_o_pt;

  DBGPrint (3, ("entering look_for_pt_schedule for targets %s PT support\n",
		(target_has_pt) ? "with" : "without"));
  
  set_analysis (input_ts, "Wang00_fixed");
  set_priorities (input_ts, INORDER);
  make_all_preemptible (input_ts);

  anneal_ts = copy_task_set (input_ts);
  greedy_ts = copy_task_set (input_ts);

  greedy_res = greedy_priorities_and_thresholds (greedy_ts, target_has_pt);
  anneal_res = anneal_priorities_and_thresholds (&anneal_ts, target_has_pt);

  DBGPrint (0, ("assigning P&PT (for target %s PT support): greedy: %s, anneal: %s\n",
		(target_has_pt) ? "with" : "without",
		(greedy_res) ? "TRUE" : "FALSE",
		(anneal_res) ? "TRUE" : "FALSE"));

  if (optimize && greedy_res) {
    double bdu;
    int threads = optimal_partition_into_threads (greedy_ts);

    DBGPrint (1, ("  *** greedy solution ***\n"));

    if (!target_has_pt) ensure_target_doesnt_need_pt (greedy_ts);

    greedy_scale_pt = find_critical_scale (greedy_ts, NULL);

    if (!target_has_pt) ensure_target_doesnt_need_pt (greedy_ts);
    assert (feasible (greedy_ts, TRUE) == num_tasks (greedy_ts));

    greedy_ts = maximize_insensitivity_by_annealing (greedy_ts, TRUE, target_has_pt);

    if (!target_has_pt) ensure_target_doesnt_need_pt (greedy_ts);
    assert (feasible (greedy_ts, TRUE) == num_tasks (greedy_ts));

    greedy_scale_o_pt = find_critical_scale (greedy_ts, NULL);

    if (!target_has_pt) ensure_target_doesnt_need_pt (greedy_ts);
    assert (feasible (greedy_ts, TRUE) == num_tasks (greedy_ts));

    bdu = greedy_scale_o_pt * utilization_set (greedy_ts);

    DBGPrint (3, ("this schedule requires %d threads and has CS = %f, BDU = %f\n", 
		  threads, greedy_scale_o_pt, bdu));

    DBGPrint (0, ("greedy solution %s PT support: orig cs = %f, opt cs = %f\n",
		  (target_has_pt) ? "with" : "without",
		  greedy_scale_pt, greedy_scale_o_pt));

  } else {
    greedy_scale_pt = -1;
    greedy_scale_o_pt = -1;
  }

  if (optimize && anneal_res) {
    double bdu;
    int threads = optimal_partition_into_threads (anneal_ts);

    DBGPrint (1, ("  *** annealing solution ***\n"));
    // print_task_set (anneal_ts);

    if (!target_has_pt) ensure_target_doesnt_need_pt (anneal_ts);
    assert (feasible (anneal_ts, TRUE) == num_tasks (anneal_ts));

    anneal_scale_pt = find_critical_scale (anneal_ts, NULL);

    if (!target_has_pt) ensure_target_doesnt_need_pt (anneal_ts);
    assert (feasible (anneal_ts, TRUE) == num_tasks (anneal_ts));

    anneal_ts = maximize_insensitivity_by_annealing (anneal_ts, TRUE, target_has_pt);

    if (!target_has_pt) ensure_target_doesnt_need_pt (anneal_ts);
    assert (feasible (anneal_ts, TRUE) == num_tasks (anneal_ts));

    anneal_scale_o_pt = find_critical_scale (anneal_ts, NULL);

    if (!target_has_pt) ensure_target_doesnt_need_pt (anneal_ts);
    assert (feasible (anneal_ts, TRUE) == num_tasks (anneal_ts));

    bdu = anneal_scale_o_pt * utilization_set (anneal_ts);

    DBGPrint (3, ("this schedule requires %d threads and has CS = %f, BDU = %f\n", 
		  threads, anneal_scale_o_pt, bdu));
	      
    DBGPrint (0, ("anneal solution %s PT support: orig cs = %f, opt cs = %f\n",
		  (target_has_pt) ? "with" : "without",
		  anneal_scale_pt, anneal_scale_o_pt));

  } else {
    anneal_scale_pt = -1;
    anneal_scale_o_pt = -1;
  }

  if (target_has_pt) {
    scale_pt_4pt = -1;
    scale_o_pt_4pt = -1;
  } else {
    scale_pt_4npt = -1;
    scale_o_pt_4npt = -1;
  }

  if (greedy_res || anneal_res) {
    if (anneal_res && (anneal_scale_o_pt >= greedy_scale_o_pt)) {
      free_task_set (greedy_ts);
      *new_ts = anneal_ts;
      
      if (target_has_pt) {
	scale_o_pt_4pt = anneal_scale_o_pt;
      } else {
	scale_o_pt_4npt = anneal_scale_o_pt;
      }

    } else {
      assert (greedy_res);
      free_task_set (anneal_ts);
      *new_ts = greedy_ts;

      if (target_has_pt) {
	scale_o_pt_4pt = greedy_scale_o_pt;
      } else {
	scale_o_pt_4npt = greedy_scale_o_pt;
      }

    }

    feas = feasible (*new_ts, TRUE);
    if (feas != num_tasks(*new_ts)) {
      print_task_set (*new_ts);
      assert (0);
    }
    DBGPrint (0, ("PT analysis: %d / %d tasks are feasible\n", 
		  feas, num_tasks (*new_ts)));

    if (!target_has_pt) ensure_target_doesnt_need_pt (*new_ts);

  } else {
    free_task_set (greedy_ts);
    free_task_set (anneal_ts);
    *new_ts = NULL;
  }
}    

static void max_robust_min_threads (struct task_set *ts,
				    int pt_support)
{
  int nthreads1, nthreads2;
  
  if (pt_support) {
    struct task_set *ts7 = copy_task_set (ts);
    nthreads1 = optimal_partition_into_threads (ts7);
    maximize_preempt_thresholds (ts7);
    nthreads2 = optimal_partition_into_threads (ts7);
    free_task_set (ts7);
  } else {
    nthreads1 = -1;
    nthreads2 = -1;
  }

  {
    struct task_set *ts8 = copy_task_set (ts);
    minimize_threads_by_annealing (ts8, test_overrun, pt_support);
  }
    
  DBGPrint (0, ("for targets %s PT support:\n", (pt_support) ? "with" : "without"));
  DBGPrint (0, ("  max insensitive set needs: %d threads, old method %d\n",
		nthreads1, nthreads2));
}

static int total_p;
static int total_np;
static int total_pt_4pt;
static int total_pt_4npt;

int preemption_thresh_test (struct task_set *ts, int num)
{
  struct task_set *ts_pt_4pt, *ts_pt_4npt;
  struct task_set *ts_np, *ts_p;
  int ret;

  test_target_no_pt = FALSE;
  test_target_yes_pt = FALSE;
  minimize = FALSE;
  optimize = TRUE;
  do_sim = FALSE;
  test_overrun = FALSE;

  DBGPrint (0, ("\n#######################################################\n"));
  DBGPrint (0, ("%d : U = %f : num tasks = %d\n", 
		num, utilization_set (ts), num_tasks(ts)));

  // print_task_set (ts);

  // save_task_set_source_code (ts);
  // save_task_set_latex (ts);

  create_clusters_for_singletons (ts);

  /*
   * preemptive scheduling
   */
  if (!has_task_clusters (ts)) {
    look_for_p_schedule (ts, &ts_p);
  } else {
    ts_p = NULL;
  }

  /*
   * non-preemptive scheduling
   */
  if (0 && !has_task_barriers (ts)) { // FIXME!

#if 0
    {
      int res1;
      struct task_set *tts = copy_task_set (ts);
      implement_clusters_using_locks (tts);
      set_priorities (tts, INORDER);
      make_all_preemptible (tts);
      set_analysis (tts, "Tindell92_restricted");
      res1 = assign_optimal_pri (tts);
      printf ("assigning priorities: %s for clusters implemented using locks\n",
	      (res1) ? "TRUE" : "FALSE");
      free_task_set (tts);
    }
#endif

    look_for_np_schedule (ts, &ts_np);

  } else {
    ts_np = NULL;
  }

  
  DBGPrint (0, ("assigning priorities:  preemptive: %s,  nonpreemptive: %s\n",
		(ts_p) ? "TRUE" : "FALSE",
		(ts_np) ? "TRUE" : "FALSE"));


  /*
   * preemption threshold scheduling
   */
  if (test_target_yes_pt) {
    // target platform has preemption thresholds
    look_for_pt_schedule (ts, &ts_pt_4pt, TRUE);
  } else {
    ts_pt_4pt = NULL;
  }
    
  if (test_target_no_pt) {
    // target platform does not have preemption thresholds
    look_for_pt_schedule (ts, &ts_pt_4npt, FALSE);
  } else {
    ts_pt_4npt = NULL;
  }

  if (ts_p) total_p++;
  if (ts_np) total_np++;
  if (ts_pt_4pt) total_pt_4pt++;
  if (ts_pt_4npt) total_pt_4npt++;

  /*
   * if possible, opportunistically grab the schedule with the
   * highest critical scaling factor
   */

  if (test_target_yes_pt) {
    
    if (ts_np && (!ts_pt_4pt || (scale_o_np > scale_o_pt_4pt))) {
      if (ts_pt_4pt) free_task_set (ts_pt_4pt);
      scale_o_pt_4pt = scale_o_np;
      ts_pt_4pt = ts_np;
      ts_np = NULL;
    }
    
    if (ts_p && (!ts_pt_4pt || (scale_o_p > scale_o_pt_4pt))) {
      if (ts_pt_4pt) free_task_set (ts_pt_4pt);
      scale_o_pt_4pt = scale_o_p;
      ts_pt_4pt = ts_p;
      ts_p = NULL;
    }

  }

  if (ts_pt_4npt) ensure_target_doesnt_need_pt (ts_pt_4npt);

  if (test_target_no_pt) {
    
    if (ts_np && (!ts_pt_4npt || (scale_o_np > scale_o_pt_4npt))) {
      if (ts_pt_4npt) free_task_set (ts_pt_4npt);
      scale_o_pt_4npt = scale_o_np;
      ts_pt_4npt = ts_np;
      ts_np = NULL;
      // put_all_tasks_in_one_cluster (ts_pt_4npt);
      DBGPrint (1, ("subst the non-preemptive task set for PT-for-no-PT\n"));
    }
    
    if (ts_p && (!ts_pt_4npt || (scale_o_p > scale_o_pt_4npt))) {
      if (ts_pt_4npt) free_task_set (ts_pt_4npt);
      scale_o_pt_4npt = scale_o_p;
      ts_pt_4npt = ts_p;
      ts_p = NULL;
      DBGPrint (1, ("subst the preemptive task set for PT-for-no-PT\n"));
    }
  }

  if (ts_pt_4npt) ensure_target_doesnt_need_pt (ts_pt_4npt);
  
  if (ts_pt_4pt && test_target_yes_pt) {
    if (do_sim) {
      simulate (ts_pt_4pt, 10000000, NULL, 0.0, NULL, stdout);
    }
    if (minimize) {
      max_robust_min_threads (ts_pt_4pt, TRUE);
    }
  }

  if (ts_pt_4npt && test_target_no_pt) {
    if (do_sim) {
      simulate (ts_pt_4npt, 10000000, NULL, 0.0, NULL, stdout);
    }
    if (minimize) {
      max_robust_min_threads (ts_pt_4npt, FALSE);
    }
  }

  if (ts_p || ts_np || ts_pt_4pt || ts_pt_4npt) {
    
    // FIXME
    assert (ts_p);
    print_task_set (ts_p);
    
    ret = 1;
  } else {
    ret = 0;
  }

  if (ts_np) free_task_set (ts_np);
  if (ts_p) free_task_set (ts_p);
  if (ts_pt_4pt) free_task_set (ts_pt_4pt);
  if (ts_pt_4npt) free_task_set (ts_pt_4npt);
  free_task_set (ts);

  return ret;
}
  
void random_preemp_thresh_tests (void)
{
  struct task_set *ts;
  int rep, total_tasks, max_deadline, REPS, pct;

  // static const char *sim_outf = "sim_output.txt";
  // static const char *sim_outf = NULL;  
  // static const int SIM_TIME = find_max_deadline (ts) * 100;
  
  total_p = 0;
  total_np = 0;
  total_pt_4pt = 0;
  total_pt_4npt = 0;

  max_deadline = 1000;
  max_resp = max_deadline * 1000; // seems sufficient...

  REPS = 500;
  // REPS = 25;
  
  assert (max_resp < (INT_MAX / 2));

#define LOW 5
#define HIGH 25
#define INC 5

  for (total_tasks = LOW; total_tasks <= HIGH; total_tasks += INC) {
    int res;
    pct = -1;
    rep = 0;
    do {
      if ((100*rep/REPS) > pct) {
	pct = 100*rep/REPS;
	DBGPrint (0, ("%d pct done (%d out of %d)\n", pct, rep, REPS));
      }
      ts = create_random_task_set (rep,
				   total_tasks, 
				   max_deadline, 
				   "Wang00_fixed",
				   1, // multiplier
				   2, // has jitter
				   TRUE // T != D
				   );

      assert (!has_task_clusters (ts));
      assert (!has_task_barriers (ts));

      res = preemption_thresh_test (ts, rep);
      
      if (res == 1) {
	printf ("countme\n");
      } else {
	printf ("dontcount\n");
      }

      rep += res;
    } while (rep<REPS);
  }

  DBGPrint (0, ("%d schedulable task sets were examined\n", REPS));

  DBGPrint (0, ("successful cases\n"));
  DBGPrint (0, ("  %d : preemptible\n", total_p));
  DBGPrint (0, ("  %d : nonpreemptible\n", total_np));
  DBGPrint (0, ("  %d : preemption threshold (for target with PT)\n", total_pt_4pt));
  DBGPrint (0, ("  %d : preemption threshold (for target without PT)\n", total_pt_4npt));
}

